import 'dart:math';

import 'package:jeweled/config/default_global_settings.dart';
import 'package:jeweled/models/game_board.dart';
import 'package:jeweled/models/game_cell.dart';
import 'package:jeweled/models/cell_location.dart';
import 'package:jeweled/models/settings_game.dart';
import 'package:jeweled/models/settings_global.dart';
import 'package:jeweled/utils/color_theme.dart';
import 'package:jeweled/utils/tools.dart';

class Game {
  final GameBoard board;
  final GameSettings gameSettings;
  final GlobalSettings globalSettings;
  List<int> shuffledColors = [];
  bool isRunning = false;
  bool isFinished = false;
  int availableBlocksCount = 0;
  int movesCount = 0;
  int score = 0;

  Game({
    required this.board,
    required this.gameSettings,
    required this.globalSettings,
    required this.shuffledColors,
    this.isRunning = false,
    this.isFinished = false,
    this.availableBlocksCount = 0,
    this.movesCount = 0,
    this.score = 0,
  });

  factory Game.createNull() {
    return Game(
      board: GameBoard.createNull(),
      gameSettings: GameSettings.createDefault(),
      globalSettings: GlobalSettings.createDefault(),
      shuffledColors: shuffleColors(DefaultGlobalSettings.defaultColorsThemeValue),
    );
  }

  factory Game.createNew({
    GameSettings? gameSettings,
    GlobalSettings? globalSettings,
  }) {
    GameSettings newGameSettings = gameSettings ?? GameSettings.createDefault();
    GlobalSettings newGlobalSettings = globalSettings ?? GlobalSettings.createDefault();

    return Game(
      board: GameBoard.createRandom(newGameSettings),
      gameSettings: newGameSettings,
      globalSettings: newGlobalSettings,
      shuffledColors: shuffleColors(newGlobalSettings.colorsTheme),
      isRunning: true,
    );
  }

  static List<int> shuffleColors(final int colorsTheme) {
    List<int> values =
        List<int>.generate(ColorTheme.getColorsCount(colorsTheme), (i) => i + 1);
    values.shuffle();

    return values;
  }

  void shuffleColorsAgain(final int colorsTheme) {
    shuffledColors = shuffleColors(colorsTheme);
  }

  GameCell getCell(CellLocation cellLocation) {
    return board.cells[cellLocation.row][cellLocation.col];
  }

  int? getCellValue(CellLocation cellLocation) {
    return getCell(cellLocation).value;
  }

  int? getCellValueShuffled(CellLocation cellLocation) {
    final int? value = getCell(cellLocation).value;
    return value != null ? shuffledColors[value - 1] : null;
  }

  void updateCellValue(CellLocation locationToUpdate, int? value) {
    board.cells[locationToUpdate.row][locationToUpdate.col].value = value;
  }

  void increaseMovesCount() {
    movesCount += 1;
  }

  void increaseScore(int? count) {
    score += (count ?? 0);
  }

  void updateAvailableBlocksCount() {
    availableBlocksCount = getAvailableBlocks(this).length;
  }

  void updateGameIsRunning(bool gameIsRunning) {
    isRunning = gameIsRunning;
  }

  void updateGameIsFinished(bool gameIsFinished) {
    isFinished = gameIsFinished;
  }

  List<CellLocation> getSiblingCells(
    final CellLocation referenceCellLocation,
    List<CellLocation> siblingCells,
  ) {
    final int boardSizeHorizontal = gameSettings.boardSize;
    final int boardSizeVertical = gameSettings.boardSize;

    final int? referenceValue = getCellValue(referenceCellLocation);

    for (int deltaRow = -1; deltaRow <= 1; deltaRow++) {
      for (int deltaCol = -1; deltaCol <= 1; deltaCol++) {
        if (deltaCol == 0 || deltaRow == 0) {
          final int candidateRow = referenceCellLocation.row + deltaRow;
          final int candidateCol = referenceCellLocation.col + deltaCol;

          if ((candidateRow >= 0 && candidateRow < boardSizeVertical) &&
              (candidateCol >= 0 && candidateCol < boardSizeHorizontal)) {
            final candidateLocation = CellLocation.go(candidateRow, candidateCol);

            if (getCellValue(candidateLocation) == referenceValue) {
              bool alreadyFound = false;
              for (int index = 0; index < siblingCells.length; index++) {
                if ((siblingCells[index].row == candidateRow) &&
                    (siblingCells[index].col == candidateCol)) {
                  alreadyFound = true;
                }
              }
              if (!alreadyFound) {
                siblingCells.add(candidateLocation);
                siblingCells = getSiblingCells(candidateLocation, siblingCells);
              }
            }
          }
        }
      }
    }

    return siblingCells;
  }

  List<List<CellLocation>> getAvailableBlocks(final Game game) {
    final int boardSizeHorizontal = game.gameSettings.boardSize;
    final int boardSizeVertical = game.gameSettings.boardSize;

    final List<List<CellLocation>> blocks = [];

    for (int row = 0; row < boardSizeVertical; row++) {
      for (int col = 0; col < boardSizeHorizontal; col++) {
        final CellLocation cellLocation = CellLocation.go(row, col);
        if (game.getCellValue(cellLocation) != null) {
          // if current cell not already in a found block
          bool alreadyFound = false;

          for (List<CellLocation> foundBlock in blocks) {
            for (CellLocation foundBlockCell in foundBlock) {
              if ((foundBlockCell.row == row) && (foundBlockCell.col == col)) {
                alreadyFound = true;
              }
            }
          }
          if (!alreadyFound) {
            final List<CellLocation> block = game.getSiblingCells(cellLocation, []);
            if (block.length >= 3) {
              blocks.add(block);
            }
          }
        }
      }
    }

    return blocks;
  }

  bool hasAtLeastOneAvailableBlock() {
    final int boardSizeHorizontal = gameSettings.boardSize;
    final int boardSizeVertical = gameSettings.boardSize;

    for (int row = 0; row < boardSizeVertical; row++) {
      for (int col = 0; col < boardSizeHorizontal; col++) {
        final CellLocation cellLocation = CellLocation.go(row, col);
        if (getCellValue(cellLocation) != null) {
          final List<CellLocation> block = getSiblingCells(cellLocation, []);
          if (block.length >= 3) {
            // found one block => ok, not locked
            return true;
          }
        }
      }
    }

    printlog('Board is locked!');
    return false;
  }

  bool isInBoard(CellLocation cell) {
    if (cell.row > 0 &&
        cell.row < gameSettings.boardSize &&
        cell.col > 0 &&
        cell.col < gameSettings.boardSize) {
      return true;
    }
    return false;
  }

  int getFillValue(CellLocation referenceCellLocation) {
    final int row = referenceCellLocation.row;
    final int col = referenceCellLocation.col;

    // build a list of values to pick one
    final List<int> values = [];

    // All eligible values (twice)
    final int maxValue = gameSettings.colorsCount;
    for (int i = 1; i <= maxValue; i++) {
      values.add(i);
      values.add(i);
    }

    // Add values of current col (twice)
    for (int r = 0; r <= gameSettings.boardSize; r++) {
      if (isInBoard(CellLocation.go(r, col))) {
        final int? value = getCellValue(CellLocation.go(r, col));
        if (value != null) {
          values.add(value);
          values.add(value);
        }
      }
    }

    // Add values of sibling cols (twice for top rows)
    for (int deltaCol = -1; deltaCol <= 1; deltaCol++) {
      final int c = col + deltaCol;
      for (int r = 0; r < gameSettings.boardSize; r++) {
        if (isInBoard(CellLocation.go(r, c))) {
          final int? value = getCellValue(CellLocation.go(r, c));
          if (value != null) {
            values.add(value);
            if (row < gameSettings.boardSize / 3) {
              values.add(value);
            }
          }
        }
      }
    }

    // Add values of sibling cells
    for (int deltaCol = -2; deltaCol <= 2; deltaCol++) {
      final int c = col + deltaCol;
      for (int deltaRow = -2; deltaRow <= 2; deltaRow++) {
        final int r = row + deltaRow;
        if (isInBoard(CellLocation.go(r, c))) {
          final int? value = getCellValue(CellLocation.go(r, c));
          if (value != null) {
            values.add(value);
          }
        }
      }
    }

    // Pick random value from "ponderated" list
    return values[Random().nextInt(values.length)];
  }

  void dump() {
    printlog('');
    printlog('## Current game dump:');
    printlog('');
    gameSettings.dump();
    globalSettings.dump();
    printlog('');
    board.dump();
    printlog('');
    printlog('Game: ');
    printlog('  isRunning: $isRunning');
    printlog('  isFinished: $isFinished');
    printlog('  movesCount: $movesCount');
    printlog('  score: $score');
    printlog('  availableBlocksCount: $availableBlocksCount');
    printlog('  shuffledColors: $shuffledColors');
    printlog('');
  }

  @override
  String toString() {
    return 'Game(${toJson()})';
  }

  Map<String, dynamic>? toJson() {
    return <String, dynamic>{
      'board': board.toJson(),
      'gameSettings': gameSettings.toJson(),
      'globalSettings': globalSettings.toJson(),
      'shuffledColors': shuffledColors,
      'isRunning': isRunning,
      'isFinished': isFinished,
      'availableBlocksCount': availableBlocksCount,
      'movesCount': movesCount,
      'score': score,
    };
  }
}
